from __future__ import annotations

from typing import Callable, Literal

from graph_pes.atomic_graph import AtomicGraph, is_batch
from graph_pes.utils.misc import random_rotation_matrix

Transform = Callable[[AtomicGraph], AtomicGraph]
TransformName = Literal["identity", "random_rotation"]


def identity_transform(graph: AtomicGraph) -> AtomicGraph:
    """A transform that returns the graph unchanged."""
    return graph


def random_rotation(graph: AtomicGraph) -> AtomicGraph:
    """Randomly rotate the atom positions, cell and forces (if present)."""

    if is_batch(graph):
        raise ValueError("Batched graphs are not supported")

    R = random_rotation_matrix()

    props = {**graph.properties}
    if "forces" in props:
        props["forces"] = props["forces"] @ R

    new_graph = graph._replace(
        R=graph.R @ R,
        cell=graph.cell @ R,
        properties=props,
    )
    return new_graph


def parse_transform(t: TransformName | Transform | None) -> Transform:
    if t is None or t == "identity":
        return identity_transform
    elif t == "random_rotation":
        return random_rotation
    elif callable(t):
        return t
    else:
        raise ValueError(f"Unknown transform: {t}")



from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from typing import Any, Final, Sequence, final

import torch
from ase.data import chemical_symbols
from torch import nn

from graph_pes.atomic_graph import AtomicGraph, PropertyKey, to_batch
from graph_pes.data.datasets import GraphDataset
from graph_pes.utils.logger import logger

from .utils.nn import PerElementParameter


class GraphPropertyModel(nn.Module, ABC):
    """
    Base class for all models that make predictions of arbitrary properties
    from graph input.

    :class:`~graph_pes.GraphPropertyModel` objects save various pieces of extra
    metadata to the ``state_dict`` via the
    :meth:`~graph_pes.GraphPropertyModel.get_extra_state` and
    :meth:`~graph_pes.GraphPropertyModel.set_extra_state` methods.
    If you want to save additional extra state to the ``state_dict`` of your
    model, please implement the
    :meth:`~graph_pes.GraphPropertyModel.extra_state`
    property and corresponding setter to ensure that you do not overwrite
    these extra metadata items.

    Parameters
    ----------
    cutoff
        The cutoff radius for the model.
    implemented_properties
        The properties that the model implements in the forward pass.
    three_body_cutoff
        The cutoff radius for this model's three-body interactions,
        if applicable.
    """

    def __init__(
        self,
        cutoff: float,
        implemented_properties: list[PropertyKey],
        three_body_cutoff: float | None,
    ):
        super().__init__()

        self._GRAPH_PES_VERSION: Final[str] = "1.0.0"

        self.cutoff: torch.Tensor
        self.register_buffer("cutoff", torch.tensor(cutoff))

        self._has_been_pre_fit: torch.Tensor
        self.register_buffer("_has_been_pre_fit", torch.tensor(0))

        self.three_body_cutoff: torch.Tensor
        self.register_buffer(
            "three_body_cutoff", torch.tensor(three_body_cutoff or 0)
        )

        self.implemented_properties = implemented_properties

    @abstractmethod
    def forward(self):
        """
        The model's forward pass. Generate all properties for the given graph
        that are in this model's ``implemented_properties`` list.
        """
        ...

    @abstractmethod
    def predict(self, graph: AtomicGraph, properties: list[PropertyKey]):
        """
        Generate (optionally batched) predictions for the given
        ``properties`` and  ``graph``.

        This method returns a dictionary mapping each requested
        ``property`` to a tensor of predictions, relying on the model's
        :meth:`~graph_pes.GraphPropertyModel.forward` implementation
        together with :func:`torch.autograd.grad` to automatically infer any
        missing properties.
        """
        pass

    # add type hints for mypy etc.
    def __call__(self, graph: AtomicGraph) -> dict[PropertyKey, torch.Tensor]:
        return super().__call__(graph)

    def non_decayable_parameters(self) -> list[torch.nn.Parameter]:
        """
        Return a list of parameters that should not be decayed during training.

        By default, this method recurses over all available sub-modules
        and calls their :meth:`non_decayable_parameters` (if it is defined).

        See :class:`~graph_pes.models.components.scaling.LocalEnergiesScaler`
        for an example of this.
        """
        found = []
        for module in self.modules():
            if module is self:
                continue
            if hasattr(module, "non_decayable_parameters"):
                found.extend(module.non_decayable_parameters())
        return found

    @torch.no_grad()
    def pre_fit_all_components(
        self,
        graphs: Sequence[AtomicGraph],
    ):
        """
        Pre-fit the model, and all its components, to the training data.

        Some models require pre-fitting to the training data to set certain
        parameters. For example, the :class:`~graph_pes.models.pairwise.LennardJones`
        model uses the distribution of interatomic distances in the training data
        to set the length-scale parameter.

        In the ``graph-pes-train`` routine, this method is called before
        "normal" training begins (you can turn this off with a config option).

        This method does two things:

        1. iterates over all the model's :class:`~torch.nn.Module` components
           (including itself) and calls their :meth:`pre_fit` method (if it exists -
           see for instance :class:`~graph_pes.models.LearnableOffset` for
           an example of a model-specific pre-fit method, and
           :class:`~graph_pes.models.components.scaling.LocalEnergiesScaler` for
           an example of a component-specific pre-fit method).
        2. registers all the unique atomic numbers in the training data with
           all of the model's :class:`~graph_pes.utils.nn.PerElementParameter`
           instances to ensure correct parameter counting.

        If the model has already been pre-fitted, subsequent calls to
        :meth:`pre_fit_all_components` will be ignored (and a warning will be raised).

        Parameters
        ----------
        graphs
            The training data.
        """  # noqa: E501

        model_name = self.__class__.__name__
        logger.debug(f"Attempting to pre-fit {model_name}")

        # 1. get the graphs as a single batch
        if isinstance(graphs, GraphDataset):
            graphs = list(graphs)
        graph_batch = to_batch(graphs)

        # 2a. if the graph has already been pre-fitted: warn
        if self._has_been_pre_fit:
            model_name = self.__class__.__name__
            logger.warning(
                f"This model ({model_name}) has already been pre-fitted. "
                "This, and any subsequent, call to pre_fit_all_components will "
                "be ignored.",
                stacklevel=2,
            )

        # 2b. if the model has not been pre-fitted: pre-fit
        else:
            if len(graphs) > 10_000:
                logger.warning(
                    f"Pre-fitting on a large dataset ({len(graphs):,} graphs). "
                    "This may take some time. Consider using a smaller, "
                    "representative collection of structures for pre-fitting. "
                    "Set ``max_n_pre_fit`` in your config, or "
                    "see GraphDataset.sample() for more information.",
                    stacklevel=2,
                )

            self.pre_fit(graph_batch)
            # pre-fit any sub-module with a pre_fit method
            for module in self.modules():
                if hasattr(module, "pre_fit") and module is not self:
                    module.pre_fit(graph_batch)

            self._has_been_pre_fit = torch.tensor(1)

        # 3. finally, register all per-element parameters (no harm in doing this
        #    multiple times)
        for param in self.parameters():
            if isinstance(param, PerElementParameter):
                param.register_elements(torch.unique(graph_batch.Z).tolist())

    def pre_fit(self, graphs: AtomicGraph) -> None:
        """
        Override this method to perform additional pre-fitting steps.

        See :class:`~graph_pes.models.components.scaling.LocalEnergiesScaler` or
        :class:`~graph_pes.models.offsets.EnergyOffset` for examples of this.

        Parameters
        ----------
        graphs
            The training data.
        """

    @torch.jit.unused
    @property
    def elements_seen(self) -> list[str]:
        """The elements that the model has seen during training."""
        Zs = set()
        for param in self.parameters():
            if isinstance(param, PerElementParameter):
                Zs.update(param._accessed_Zs)
        return [chemical_symbols[Z] for Z in sorted(Zs)]

    @torch.jit.unused
    @property
    def device(self) -> torch.device:
        return self.cutoff.device

    @torch.jit.unused
    @final
    def get_extra_state(self) -> dict[str, Any]:
        """
        Get the extra state of this instance. Please override the
        :meth:`~graph_pes.GraphPropertyModel.extra_state` property to add extra
        state here.
        """
        return {
            "_GRAPH_PES_VERSION": self._GRAPH_PES_VERSION,
            "extra": self.extra_state,
        }

    @torch.jit.unused
    @final
    def set_extra_state(self, state: dict[str, Any]) -> None:  # type: ignore
        """
        Set the extra state of this instance using a dictionary mapping strings
        to values returned by the
        :meth:`~graph_pes.GraphPropertyModel.extra_state` property setter to add
        extra state here.
        """
        version = state.pop("_GRAPH_PES_VERSION", None)
        if version is not None:
            current_version = self._GRAPH_PES_VERSION
            if version != current_version:
                warnings.warn(
                    "You are attempting to load a state dict corresponding "
                    f"to graph-pes version {version}, but the current version "
                    f"of this model is {current_version}. This may cause "
                    "errors when loading the model.",
                    stacklevel=2,
                )
            self._GRAPH_PES_VERSION = version  # type: ignore

        # user defined extra state
        self.extra_state = state["extra"]

    @torch.jit.unused
    @property
    def extra_state(self) -> Any:
        """
        Override this property to add extra state to the model's
        ``state_dict``.
        """
        return {}

    @torch.jit.unused
    @extra_state.setter
    def extra_state(self, state: Any) -> None:
        """
        Set the extra state of this instance using a value returned by the
        :meth:`~graph_pes.GraphPESModel.extra_state` property.
        """
        pass


class GraphTensorModel(GraphPropertyModel):
    r"""
    Base class for all models that make predictions of
    arbitrary rank, per-atom tensors from graph input.

    Parameters
    ----------
    cutoff
        The cutoff radius for the model.
    implemented_properties
        The property predictions that the model implements in the forward pass.
        Must include ``"tensor"``.
    """

    def __init__(
        self,
        cutoff: float,
        implemented_properties: list[PropertyKey],
    ):
        super().__init__(
            cutoff=cutoff,
            implemented_properties=implemented_properties,
            three_body_cutoff=None,
        )

    @abstractmethod
    def forward(self):
        pass

    def predict(self, graph: AtomicGraph, properties: list[PropertyKey]):
        predictions = self(graph)
        predictions = {prop: predictions[prop] for prop in properties}

        # tidy up if in eval mode
        if not self.training:
            predictions = {k: v.detach() for k, v in predictions.items()}

        return predictions



from __future__ import annotations

from pytorch_lightning.loggers import Logger as PTLLogger

from graph_pes.atomic_graph import (
    AtomicGraph,
    number_of_atoms,
    number_of_structures,
)
from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.graph_property_model import GraphPropertyModel, GraphTensorModel
from graph_pes.models.addition import AdditionModel, TensorAdditionModel
from graph_pes.utils.logger import logger
from graph_pes.utils.nn import count_used_parameters


def log_model_info(
    model: GraphPropertyModel,
    ptl_logger: PTLLogger | None = None,
) -> None:
    """Log the number of parameters in a model."""

    logger.debug(f"Model:\n{model}")

    if isinstance(model, (AdditionModel, TensorAdditionModel)):
        model_names = [
            f"{given_name} ({component.__class__.__name__})"
            for given_name, component in model.models.items()
        ]
        params = [
            count_used_parameters(component, only_learnable=True)
            for component in model.models.values()
        ]
        width = max(len(name) for name in model_names)
        info_str = "Number of learnable params:"
        for name, param in zip(model_names, params):
            info_str += f"\n    {name:<{width}}: {param:,}"
        logger.info(info_str)

    else:
        n = count_used_parameters(model, only_learnable=True)
        logger.info(f"Number of learnable params : {n:,}")

    if ptl_logger is not None:
        all_params = count_used_parameters(model, only_learnable=False)
        learnable_params = count_used_parameters(model, only_learnable=True)
        ptl_logger.log_metrics(
            {
                "n_parameters": all_params,
                "n_learnable_parameters": learnable_params,
            }
        )


def sanity_check(model: GraphPropertyModel, batch: AtomicGraph) -> None:
    if isinstance(model, GraphPESModel):
        outputs = model.get_all_PES_predictions(batch)

        N = number_of_atoms(batch)
        S = number_of_structures(batch)
        expected_shapes = {
            "local_energies": (N,),
            "forces": (N, 3),
            "energy": (S,),
            "stress": (S, 3, 3),
            "virial": (S, 3, 3),
        }

        incorrect = []
        for key, value in outputs.items():
            if value.shape != expected_shapes[key]:
                incorrect.append((key, value.shape, expected_shapes[key]))

        if len(incorrect) > 0:
            raise ValueError(
                "Sanity check failed for the following outputs:\n"
                + "\n".join(
                    f"{key}: {value} != {expected}"
                    for key, value, expected in incorrect
                )
            )

        if batch.cutoff < model.cutoff:
            logger.error(
                "Sanity check failed: you appear to be training on data "
                f"composed of graphs with a cutoff ({batch.cutoff}) that is "
                f"smaller than the cutoff used in the model ({model.cutoff}). "
                "This is almost certainly not what you want to do?",
            )

    elif isinstance(model, GraphTensorModel):
        outputs = model.predict(batch, ["tensor"])

        # for now, only a single output, "tensor", is supported
        if "tensor" not in outputs:
            raise ValueError(
                "Sanity check failed: the model did not predict the "
                "`tensor` property."
            )

        tensor = outputs["tensor"]
        N = number_of_atoms(batch)
        if tensor.shape[0] != N:
            raise ValueError(
                "Sanity check failed: the model predicted a tensor with "
                f"shape {tensor.shape} but the number of atoms in the graph "
                f"is {N}."
            )


VALIDATION_LOSS_KEY = "valid/loss/total"



from __future__ import annotations

from typing import Literal

import torch
from torch import Tensor

from graph_pes.atomic_graph import (
    DEFAULT_CUTOFF,
    AtomicGraph,
    PropertyKey,
    edge_wise_softmax,
    keep_at_most_k_neighbours,
    neighbour_distances,
    neighbour_vectors,
    remove_mean_and_net_torque,
    sum_over_central_atom_index,
)
from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.models.components.distances import Bessel, PolynomialEnvelope
from graph_pes.models.components.scaling import LocalEnergiesScaler
from graph_pes.utils.nn import (
    MLP,
    PerElementEmbedding,
    ShiftedSoftplus,
    UniformModuleList,
)

from .e3nn.utils import SphericalHarmonics

# TODO: penalise rotational grad


NormType = Literal["layer", "rms"]
AttentionGate = Literal["sigmoid", "softmax"]


def get_norm(norm_type: NormType):
    if norm_type == "layer":
        return torch.nn.LayerNorm
    elif norm_type == "rms":
        return torch.nn.RMSNorm
    else:
        raise ValueError(f"Unknown norm type: {norm_type}")


class OrbEncoder(torch.nn.Module):
    """Generates node and edge features embeddings for an atomic graph."""

    def __init__(
        self,
        cutoff: float,
        channels: int,
        radial_features: int,
        l_max: int,
        edge_outer_product: bool,
        mlp_layers: int,
        mlp_hidden_dim: int,
        activation: str,
        norm_type: NormType,
    ):
        super().__init__()
        self.cutoff = cutoff
        self.channels = channels
        self.edge_outer_product = edge_outer_product

        # nodes
        self.Z_embedding = PerElementEmbedding(channels)
        self.Z_layer_norm = torch.nn.LayerNorm(channels)

        # edges
        self.rbf = Bessel(radial_features, cutoff, trainable=False)
        self.envelope = PolynomialEnvelope(p=4, cutoff=cutoff)
        self.sh = SphericalHarmonics(
            [l for l in range(l_max + 1)],
            normalize=True,
            normalization="component",
        )
        sh_dim: int = self.sh.irreps_out.dim  # type: ignore
        self.edge_dim = (
            radial_features * sh_dim
            if edge_outer_product
            else radial_features + sh_dim
        )
        self.edge_mlp = MLP(
            [self.edge_dim] + [mlp_hidden_dim] * mlp_layers + [channels],
            activation,
        )
        self.edge_layer_norm = get_norm(norm_type)(channels)

    def forward(self, graph: AtomicGraph) -> tuple[Tensor, Tensor]:
        node_emb = self.Z_layer_norm(self.Z_embedding(graph.Z))

        # featurise angles
        v = neighbour_vectors(graph)
        sh_emb = self.sh(v)

        # featurise distances
        d = torch.linalg.norm(v, dim=-1)
        rbf_emb = self.rbf(d)

        # combine
        if self.edge_outer_product:
            edge_emb = rbf_emb[:, :, None] * sh_emb[:, None, :]
        else:
            edge_emb = torch.cat([rbf_emb, sh_emb], dim=1)
        edge_emb = edge_emb.view(-1, self.edge_dim)

        # smooth cutoff
        c = self.envelope(d)
        edge_feats = edge_emb * c.unsqueeze(-1)

        # mlp
        edge_emb = self.edge_layer_norm(self.edge_mlp(edge_feats))

        return node_emb, edge_emb


class OrbMessagePassingLayer(torch.nn.Module):
    def __init__(
        self,
        cutoff: float,
        channels: int,
        mlp_layers: int,
        mlp_hidden_dim: int,
        activation: str,
        norm_type: NormType,
        attention_gate: AttentionGate,
        distance_smoothing: bool,
    ):
        super().__init__()

        self.node_mlp = torch.nn.Sequential(
            MLP(
                [channels * 3] + [mlp_hidden_dim] * mlp_layers + [channels],
                activation,
            ),
            get_norm(norm_type)(channels),
        )

        self.edge_mlp = torch.nn.Sequential(
            MLP(
                [channels * 3] + [mlp_hidden_dim] * mlp_layers + [channels],
                activation,
            ),
            get_norm(norm_type)(channels),
        )

        self.receive_attn = torch.nn.Linear(channels, 1)
        self.send_attn = torch.nn.Linear(channels, 1)

        self.attention_gate = attention_gate

        if distance_smoothing:
            self.envelope = PolynomialEnvelope(p=4, cutoff=cutoff)
        else:
            self.envelope = None

    def forward(
        self,
        node_emb: Tensor,  # (N, C)
        edge_emb: Tensor,  # (E, C)
        graph: AtomicGraph,
    ) -> tuple[Tensor, Tensor]:
        # calculate per-edge attention weights based on both
        # senders and receivers
        if self.attention_gate == "softmax":
            receive_attn_weights = edge_wise_softmax(
                self.receive_attn(edge_emb), graph, aggregation="receivers"
            )
            send_attn_weights = edge_wise_softmax(
                self.send_attn(edge_emb), graph, aggregation="senders"
            )
        elif self.attention_gate == "sigmoid":
            receive_attn_weights = torch.sigmoid(self.receive_attn(edge_emb))
            send_attn_weights = torch.sigmoid(self.send_attn(edge_emb))
        else:
            raise ValueError(f"Unknown attention gate: {self.attention_gate}")

        # optionally decay these weights near the cutoff
        if self.envelope is not None:
            envelope = self.envelope(neighbour_distances(graph)).unsqueeze(-1)
            receive_attn_weights = receive_attn_weights * envelope
            send_attn_weights = send_attn_weights * envelope

        # generate new edge features
        new_edge_features = torch.cat(
            [
                edge_emb,
                node_emb[graph.neighbour_list[0]],
                node_emb[graph.neighbour_list[1]],
            ],
            dim=1,
        )
        new_edge_features = self.edge_mlp(new_edge_features)

        #  generate new node features from attention weights
        senders, receivers = graph.neighbour_list[0], graph.neighbour_list[1]
        sent_total_message = sum_over_central_atom_index(  # (N, C)
            new_edge_features * send_attn_weights, senders, graph
        )
        received_total_message = sum_over_central_atom_index(  # (N, C)
            new_edge_features * receive_attn_weights, receivers, graph
        )
        new_node_features = torch.cat(
            [node_emb, sent_total_message, received_total_message],
            dim=1,
        )
        new_node_features = self.node_mlp(new_node_features)

        # residual connection
        node_emb = node_emb + new_node_features
        edge_emb = edge_emb + new_edge_features

        return node_emb, edge_emb


class Orb(GraphPESModel):
    r"""
    The `Orb-v3 <https://arxiv.org/abs/2504.06231>`__ architecture.

    Citation:

    .. code-block:: bibtex

        @misc{Rhodes-25-04,
            title = {Orb-v3: Atomistic Simulation at Scale},
            author = {
                Rhodes, Benjamin and Vandenhaute, Sander
                and {\v S}imkus, Vaidotas and Gin, James and Godwin, Jonathan
                and Duignan, Tim and Neumann, Mark
            },
            year = {2025},
            publisher = {arXiv},
            doi = {10.48550/arXiv.2504.06231},
        }

    Parameters
    ----------
    cutoff
        The cutoff radius for interatomic interactions.
    conservative
        If ``True``, the model will generate force predictions as the negative
        gradient of the energy with respect to atomic positions. If ``False``,
        the model will have a separate force prediction head.
    channels
        The number of channels in the model.
    layers
        The number of message passing layers.
    radial_features
        The number of radial basis functions to use.
    mlp_layers
        The number of layers in the MLPs.
    mlp_hidden_dim
        The hidden dimension of the MLPs.
    l_max
        The maximum degree of spherical harmonics to use.
    edge_outer_product
        If ``True``, use the outer product of radial and angular features for
        edge embeddings. If ``False``, concatenate radial and angular features.
    activation
        The activation function to use in the MLPs.
    norm_type
        The type of normalization to use in the MLPs. Either ``"layer"`` for
        :class:`torch.nn.LayerNorm` or ``"rms"`` for :class:`torch.nn.RMSNorm`.
    attention_gate
        The type of attention gating to use in message passing layers. Either
        ``"sigmoid"`` for element-wise sigmoid gating or ``"softmax"`` for
        normalising attention weights over neighbours.
    distance_smoothing
        If ``True``, apply a polynomial envelope to attention weights based on
        interatomic distances. If ``False``, do not apply any distance-based
        smoothing.
    max_neighbours
        If set, limit the number of neighbours per atom to this value by
        keeping only the closest ones.
    """

    def __init__(
        self,
        cutoff: float = DEFAULT_CUTOFF,
        conservative: bool = False,
        channels: int = 256,
        layers: int = 5,
        radial_features: int = 8,
        mlp_layers: int = 2,
        mlp_hidden_dim: int = 1024,
        l_max: int = 3,
        edge_outer_product: bool = True,
        activation: str = "silu",
        norm_type: NormType = "layer",
        attention_gate: AttentionGate = "sigmoid",
        distance_smoothing: bool = True,
        max_neighbours: int | None = None,
    ):
        props: list[PropertyKey] = (
            ["local_energies"] if conservative else ["local_energies", "forces"]
        )
        super().__init__(implemented_properties=props, cutoff=cutoff)

        self.max_neighbours = max_neighbours

        # backbone
        self._encoder = OrbEncoder(
            cutoff=cutoff,
            channels=channels,
            radial_features=radial_features,
            l_max=l_max,
            edge_outer_product=edge_outer_product,
            mlp_layers=mlp_layers,
            mlp_hidden_dim=mlp_hidden_dim,
            activation=activation,
            norm_type=norm_type,
        )
        self._gnn_layers = UniformModuleList(
            [
                OrbMessagePassingLayer(
                    channels=channels,
                    mlp_layers=mlp_layers,
                    mlp_hidden_dim=mlp_hidden_dim,
                    activation=activation,
                    norm_type=norm_type,
                    attention_gate=attention_gate,
                    distance_smoothing=distance_smoothing,
                    cutoff=cutoff,
                )
                for _ in range(layers)
            ]
        )

        # readouts
        self._energy_readout = MLP(
            [channels] + [mlp_hidden_dim] * mlp_layers + [1],
            activation=ShiftedSoftplus(),
        )
        self.scaler = LocalEnergiesScaler()
        if conservative:
            self._force_readout = None
        else:
            self._force_readout = MLP(
                [channels] + [mlp_hidden_dim] * mlp_layers + [3],
                activation=ShiftedSoftplus(),
                bias=False,
            )

    def forward(self, graph: AtomicGraph) -> dict[PropertyKey, Tensor]:
        if self.max_neighbours is not None:
            graph = keep_at_most_k_neighbours(graph, self.max_neighbours)

        # embed the graph
        node_emb, edge_emb = self._encoder(graph)

        # message passing
        for layer in self._gnn_layers:
            node_emb, edge_emb = layer(node_emb, edge_emb, graph)

        # readout
        raw_energies = self._energy_readout(node_emb)
        preds: dict[PropertyKey, Tensor] = {
            "local_energies": self.scaler(raw_energies, graph)
        }

        if self._force_readout is not None:
            raw_forces = self._force_readout(node_emb)
            preds["forces"] = remove_mean_and_net_torque(raw_forces, graph)

        return preds



from __future__ import annotations

import logging
from typing import Final

import pytorch_lightning as pl


def silently_create_trainer(**trainer_kwargs) -> pl.Trainer:
    logger = logging.getLogger("pytorch_lightning.utilities.rank_zero")
    logger.setLevel(logging.ERROR)
    _trainer = pl.Trainer(**trainer_kwargs)
    logger.setLevel(logging.INFO)
    return _trainer


# dirty hack: just get lightning to work this out,
# and ensure no annoying printing happens
_trainer = silently_create_trainer(logger=False, devices="auto")
GLOBAL_RANK: Final[int] = _trainer.global_rank
WORLD_SIZE: Final[int] = _trainer.world_size
IS_RANK_0: Final[bool] = GLOBAL_RANK == 0
